# -*- coding: utf-8 -*-

# Commented out IPython magic to ensure Python compatibility.
# %matplotlib inline
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib import cm
import math
class Error(Exception):
    """Base class for exceptions in this module."""
    pass
class Viz():
    f = None
    X  = None
    Y = None
    Z = None
    data = None

    def __init__(self,func,lb,ub,dim,func_args=None):
        """
        Args:
            func: the function to be evaluated for plotting.
            dim: the number of dimensions of solution
            lb: list containing lower bounds for each dimension
            ub: list containing upper bounds for each dimension
            func_args: args to be passed to evaluation function

        """
        self.f = func
        self.dim = dim
        self.func_args = func_args
        #self.data  = data #data should be numpy array of shape(2,)
        self.lb = lb
        self.ub = ub
        #self.pop_hist = pop_hist
    def set(self,dim,data,fitness,best,middle,worst):
        """set variables for plotting
        Args:
            dim: the number of dimensions of solution
            data: the best value returned by getstatistics function in GSK
            fitness: fitness values history
            middle: middle indviduals history
            worst: worst indviduals history
        """
        self.dim = dim
        self.data = data
        self.best_hist = best
        self.middle_hist = middle
        self.worst_hist = worst
        self.fitness = fitness
        x = np.linspace(self.lb, self.ub,  30)
        y = np.linspace(self.lb, self.ub, 30)
        self.X, self.Y = np.meshgrid(x, y)
        self.Z = self.evaluateFun(self.X,self.Y)
    def evaluateFun(self,X,Y):
        #print('shape',X.shape)
        Z =  np.zeros((X.shape[0],X.shape[1]))
        for i in range(X.shape[0]):
            for j in range(X.shape[1]):
                hold = np.array([X[i,j],Y[i,j]]).reshape(1,2)
                #print(self.dim,self.func_num)
                Z[i,j] = self.f(hold,self.func_args)
        return Z

    def update_best(self,num,data,line):

        x = self.data[num][0]
        y = self.data[num][1]
        #print(x,y)
        #print (num,x,y,self.f([x,y]))
        line.set_data(x,y)
        line.set_3d_properties(self.f(np.array(self.data[num]).reshape(1,self.dim),self.func_args)[0], 'z')
        return line,



    def build_plot(self,fig_size=(10,5),save=None, show=True):
        """build an animated plot of GSK search in action
        Args:
            fig_size: fig_size for pyplot
            save: file name to save the animation
            show: show the animated figure in matplotlib window
        """
        if len(self.best_hist) <=0 or len(self.middle_hist) <=0 or len(self.worst_hist) <=0:
            raise Error('History is empty, make sure you set track=True in run() function')
        fig = plt.figure(figsize=fig_size)
        ax = plt.axes(projection='3d')
        ax.contour3D(self.X, self.Y, self.Z, 50, cmap=cm.cool,alpha=0.3)
        ax.view_init(60, 35)
        l, = plt.plot([], [], markerfacecolor='g', markeredgecolor='k', marker='o', label='best value', markersize=30, alpha=0.3)

        lines = [plt.plot([], [], markerfacecolor='b', markeredgecolor='b', marker='o', markersize=5, alpha=0.5)[0] for _ in range(self.best_hist[0].shape[0])]
        lines2 = [plt.plot([], [], markerfacecolor='m', markeredgecolor='m', marker='|', markersize=9, alpha=0.5)[0] for _ in range(self.middle_hist[0].shape[0])]
        lines3 = [plt.plot([], [], markerfacecolor='y', markeredgecolor='y', marker='x',markersize=5, alpha=0.5)[0] for _ in range(self.worst_hist[0].shape[0])]
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        ax.legend([l ,lines[0],lines2[0],lines3[0]], ['best_value','better pepole','middle pepole','worst pepole'])
        '''
        def update_pop(num):
            for j,line in enumerate(lines):
                #if j < self.pop_hist[num].shape[0]:
                x = self.pop_hist[num][j,0]
                y = self.pop_hist[num][j,1]
                #print(num,j,x,y)

                line.set_data(x,y)
                #print(self.f(np.array([x,y]).reshape(1,2),self.dim,self.func_num))

                line.set_3d_properties(self.f(np.array(self.pop_hist[num][j]).reshape(1,self.dim),[self.dim,self.func_num])[0], 'z')

            return lines
        '''
        def update(num,data,line):
            x = self.data[num][0]
            y = self.data[num][1]
            #print(x,y)
            #print (num,x,y,self.f([x,y]))
            line.set_data(x,y)
            line.set_3d_properties(self.f(np.array(self.data[num]).reshape(1,self.dim),self.func_args)[0], 'z')


            for j,l in enumerate(lines):
                #if j < self.pop_hist[num].shape[0]:
                x = self.best_hist[num][j,0]
                y = self.best_hist[num][j,1]
                #print(num,j,x,y)

                l.set_data(x,y)
                #print(self.f(np.array([x,y]).reshape(1,2),self.dim,self.func_num))

                l.set_3d_properties(self.fitness[num][0][j], 'z')
            for j,l in enumerate(lines2):
                #if j < self.pop_hist[num].shape[0]:
                x = self.middle_hist[num][j,0]
                y = self.middle_hist[num][j,1]
                #print(num,j,x,y)

                l.set_data(x,y)
                #print(self.f(np.array([x,y]).reshape(1,2),self.dim,self.func_num))

                l.set_3d_properties(self.fitness[num][1][j], 'z')
            for j,l in enumerate(lines3):
                #if j < self.pop_hist[num].shape[0]:
                x = self.worst_hist[num][j,0]
                y = self.worst_hist[num][j,1]
                #print(num,j,x,y)

                l.set_data(x,y)
                #print(self.f(np.array([x,y]).reshape(1,2),self.dim,self.func_num))

                l.set_3d_properties(self.fitness[num][2][j], 'z')


            return (line,lines,lines2,lines3)
        def init():
            #init lines
            for line in lines:
                line.set_data([], [])
            for line in lines2:
                line.set_data([], [])
            for line in lines3:
                line.set_data([], [])


            return (lines,lines2,lines3) #return everything that must be updated

        #best_indv_ani = animation.FuncAnimation(fig, self.update_best, self.data.shape[0], fargs=(self.data,l),
        #                                   interval=100, blit=False)
        ani3 = animation.FuncAnimation(fig, update, len(self.best_hist),init_func=init, frames= 100,fargs=(self.data,l),
                                                                           interval=100, blit=False)
        if save != None:
            ani3.save(save+'.mp4')
        #lines_ani = animation.FuncAnimation(fig, update_pop, frames=len(self.pop_hist), init_func=init,
        #                                   interval=100, blit=False)
        if show:
            plt.show()
        return fig, ani3
    def plot_losses(self,losses):
        """plot animated figure of the objective value
        Args:
            losses: history of objective values from getstatistics function in GSK
        """
        fig = plt.figure()
        ax = plt.axes(xlim=(0, 100), ylim=(-100, 0))
        N = 2
        color = ['g', 'b']
        lines = [plt.plot([], [],color[c])[0] for c in range(N)]
        def init():
            #init lines
            for line in lines:
                line.set_data([], [])

            return lines

        def animate(i):

            #animate lines
            for j,line in enumerate(lines):
                line.set_data([i,i],[i,losses[j][i]])



            return lines

        anim = animation.FuncAnimation(fig, animate, init_func=init,
                                       frames=len(losses[0]), interval=100, blit=True)

        plt.show()



import random
import numpy as np

def Gained_Shared_Junior_R1R2R3(indBest=None):

    pop_size=len(indBest)
# Gained_Shared_Junior_R1R2R3.m:4
    #R0=range(0,pop_size)
# Gained_Shared_Junior_R1R2R3.m:5
    R1=[None for i in range(pop_size)]
# Gained_Shared_Junior_R1R2R3.m:6
    R2=[None for i in range(pop_size)]
# Gained_Shared_Junior_R1R2R3.m:7
    R3=[None for i in range(pop_size)]
# Gained_Shared_Junior_R1R2R3.m:8
    for ind, i in enumerate(indBest):
        #ind=find(indBest == i)
# Gained_Shared_Junior_R1R2R3.m:11
        if ind == 0:
            R1[i]=indBest[1]
# Gained_Shared_Junior_R1R2R3.m:13
            R2[i]=indBest[2]
# Gained_Shared_Junior_R1R2R3.m:14
        else:
            if ind == (pop_size-1):
                R1[i]=indBest[pop_size - 3]
# Gained_Shared_Junior_R1R2R3.m:16
                R2[i]=indBest[pop_size - 2]
# Gained_Shared_Junior_R1R2R3.m:17
            else:
                R1[i]=indBest[ind - 1]
# Gained_Shared_Junior_R1R2R3.m:19
                R2[i]=indBest[ind + 1]
# Gained_Shared_Junior_R1R2R3.m:20

        R3[i]= random.choice([x for x in range(pop_size) if x != R1[i] and x != R2[i] and x != i]) #floor(dot(rand(1,pop_size),pop_size)) + 1
# Gained_Shared_Junior_R1R2R3.m:24
    #for i in arange(1,99999999).reshape(-1):
        #pos=(logical_or(logical_or((R3 == R2),(R3 == R1)),(R3 == R0)))
# Gained_Shared_Junior_R1R2R3.m:27
        #if sum(pos) == 0:
            #break
        #else:
            #R3[pos]=floor(dot(rand(1,sum(pos)),pop_size)) + 1
# Gained_Shared_Junior_R1R2R3.m:31
        #if i > 1000:
            #error('Can not genrate R3 in 1000 iterations')
    return np.array(R1),np.array(R2),np.array(R3)



from ctypes import CDLL, POINTER, c_int, c_double
import os
import numpy as np


def cec17_test_func(x,args):
    dll_path=CDLL(os.path.abspath('lib/cec17_test_func.so'))
    f=[]
    #f_pop = []

    nx=args[0]
    mx=1

    func_num=args[1]
    functions = dll_path
    x_pointer_type = POINTER(c_double * nx)
    f_pointer_type = POINTER(c_double * mx)
    nx_type = c_int
    mx_type = c_int
    func_num_type = c_int
    functions.cec17_test_func.argtypes = [x_pointer_type, f_pointer_type,
                                          nx_type, mx_type, func_num_type]
    functions.cec17_test_func.restype = None
    x_ctype = (c_double * nx)()
    for _, indv in enumerate(x):
        for i, value in enumerate(indv):
            if len(x.shape) > 1:
                x_ctype[i] = value
            else:
                x_ctype[i] = value
        f_ctype = (c_double * mx)()
        for i in range(mx):
            f_ctype[i] = 0
        functions.cec17_test_func(x_pointer_type(x_ctype), f_pointer_type(f_ctype),
                                  nx, mx, func_num)
        #for i in range(len(f)):
        f.append(f_ctype[0])
        #f_pop.append(f)
    return np.array(f)
def cec20_test_func(x,nx=10,func_num=1):
    dll_path=CDLL(os.path.abspath('cec20_test_func.so'))
    f=[]
    mx=1
    nx=nx
    functions = dll_path
    x_pointer_type = POINTER(c_double * nx)
    f_pointer_type = POINTER(c_double * mx)
    nx_type = c_int
    mx_type = c_int
    func_num_type = c_int
    functions.cec20_test_func.argtypes = [x_pointer_type, f_pointer_type,
                                          nx_type, mx_type, func_num_type]
    functions.cec20_test_func.restype = None

    x_ctype = (c_double * nx)()
    for _, indv in enumerate(x):
        for i, value in enumerate(indv):
            if len(x.shape) > 1:
                x_ctype[i] = value
            else:
                x_ctype[i] = value
        f_ctype = (c_double * mx)()
        for i in range(mx):
            f_ctype[i] = 0
        functions.cec20_test_func(x_pointer_type(x_ctype), f_pointer_type(f_ctype),
                                  nx, mx, func_num)
        #for i in range(len(f)):
        f.append(f_ctype[0])
        #f_pop.append(f)
    return np.array(f)



import random
import numpy as np

def Gained_Shared_Middle_R1R2R3(pop_size):


# Gained_Shared_Junior_R1R2R3.m:4
    #R0=range(0,pop_size)
# Gained_Shared_Junior_R1R2R3.m:5
    R1=[None for i in range(pop_size)]
# Gained_Shared_Junior_R1R2R3.m:6
    R2=[None for i in range(pop_size)]
# Gained_Shared_Junior_R1R2R3.m:7
    R3=[None for i in range(pop_size)]
    R4=[None for i in range(pop_size)]
    R5=[None for i in range(pop_size)]
    R6=[None for i in range(pop_size)]
# Gained_Shared_Junior_R1R2R3.m:8
    range_idx = pop_size
    for  i in range(pop_size):
        #R1[i] = ind

        R1[i]= random.choice([x for x in range(range_idx) if x != i])
        R2[i]= random.choice([x for x in range(range_idx) if x != R1[i] and x != i])
        R3[i]= random.choice([x for x in range(range_idx) if x != R2[i] and x != R1[i] and x != i])
        R4[i]= random.choice([x for x in range(range_idx) if x != R2[i] and x != R1[i] and x != R3[i] and x != i])
        R5[i]= random.choice([x for x in range(range_idx) if x != R2[i] and x != R1[i] and x != R3[i] and x != R4[i] and x != i])
        R6[i]= random.choice([x for x in range(range_idx) if x != R2[i] and x != R1[i] and x != R3[i] and x != R4[i] and x != R5[i] and x != i])
        #R4[i] = ind

         #floor(dot(rand(1,pop_size),pop_size)) + 1
# Gained_Shared_Junior_R1R2R3.m:24
    #for i in arange(1,99999999).reshape(-1):
        #pos=(logical_or(logical_or((R3 == R2),(R3 == R1)),(R3 == R0)))
# Gained_Shared_Junior_R1R2R3.m:27
        #if sum(pos) == 0:
            #break
        #else:
            #R3[pos]=floor(dot(rand(1,sum(pos)),pop_size)) + 1
# Gained_Shared_Junior_R1R2R3.m:31
        #if i > 1000:
            #error('Can not genrate R3 in 1000 iterations')
    return np.array(R1),np.array(R2),np.array(R3),np.array(R4),np.array(R5),np.array(R6)



class Vectorize(object):
    def __init__ (self,DataFrame):
        self.DataFrame = DataFrame
    def Vectorize(num_vectors, vectors_keys):
        #vectors_keys = [['col1', 'col2'],[col1]]
        vectors_list = []
        for i in range(num_vectors):
            vector = []
        for row in self.DataFrame.DataFrame:
            vector.append()



import random
import numpy as np

# Hazem used 0.2 and 0.8 instead of 0.1 and 0.9 to enable population size of 5
def Gained_Shared_Senior_R1R2R3(indBest=None,p=0.1):

    pop_size=len(indBest)
# Gained_Shared_Senior_R1R2R3.m:4
    #R1=indBest(arange(1,round(dot(pop_size,0.1))))
# Gained_Shared_Senior_R1R2R3.m:6
    #R1rand=ceil(dot(length(R1),rand(pop_size,1)))
# Gained_Shared_Senior_R1R2R3.m:7
    #R1=R1(R1rand)
    R1=[indBest[random.randint(0, round(pop_size*p)-1)] for x in range(pop_size)]
# Gained_Shared_Senior_R1R2R3.m:8
    #R2=indBest(arange(round(dot(pop_size,0.1)) + 1,round(dot(pop_size,0.9))))
# Gained_Shared_Senior_R1R2R3.m:10
    #R2rand=ceil(dot(length(R2),rand(pop_size,1)))
# Gained_Shared_Senior_R1R2R3.m:11
    #R2=R2(R2rand)
    R2=[indBest[random.randint(round(pop_size*p), round(pop_size*(1-p))-1)] for x in range(pop_size)]

# Gained_Shared_Senior_R1R2R3.m:12
    #R3=indBest(arange(round(dot(pop_size,0.9)) + 1,end()))
# Gained_Shared_Senior_R1R2R3.m:14
    #R3rand=ceil(dot(length(R3),rand(pop_size,1)))
# Gained_Shared_Senior_R1R2R3.m:15
    #R3=R3(R3rand)
    R3=[indBest[random.randint(round(pop_size*(1-p)), pop_size-1)] for x in range(pop_size)]
# Gained_Shared_Senior_R1R2R3.m:16
    return np.array(R1),np.array(R2),np.array(R3)
